#include "common.hh"
#include "debug.hh"

#include "Symbol.hh"
#include "Token.hh"

#include <set>
#include <vector>

Symbol * const Symbol::SKIP_SYMBOL = new Symbol();
Symbol * const Symbol::EPSILON_SYMBOL = new Symbol();

string stringify(const vector<string> & v)
{
  if(v.size()==0) return "";
  if(v.size()==1) return v[0];
  string s = "{";
  int n = 0;
  for(vector<string>::const_iterator i=v.begin(), z=v.end(); 
      i!=z; i++)
    {
      string str  = *i;
      s += (n ? "," : "") + str;
      n++;
    }
  s += "}";
  return s;
}

string stringify(set<string> & v)
{
  if(v.size()==0) return "";
  if(v.size()==1) return *(v.begin());
  string s = "{";
  int n = 0;
  for(set<string>::const_iterator i=v.begin(), z=v.end();
      i!=z; i++)
    {
      string str  = *i;
      s += (n ? "," : "") + str;
      n++;
    }
  return s + "}";
}

string TagField::toString() const
{
  string s = 
    stringify(*tags) +
    (rewriter!="" ? "[->"+rewriter+"]" : "");
  return s;
}
bool TagField::contains(string tag) const
{
  bool b = (tags->find(tag) != tags->end());
  return b;
}
string TagField::first() const
{
  if(tags->size()==0) return "";
  return *(tags->begin());
}

string Symbol::toString() const
{
  string s =
    (
     funcs->toString() + (forms->toString() !="" ? ":" : "") + 
     forms->toString()+
     ((sfuncs->toString()!=""||sforms->toString()!="") ?"_" : "") +
     sfuncs->toString()+(sforms->toString()!="" ? ":" : "" ) +
     sforms->toString()+
     (baseforms->toString()!="" ? "\""+baseforms->toString()+"\"" : "")+
     (wordforms->toString()!="" ? "\'"+wordforms->toString()+"\'" : "")
     );
  return s;
}

//
// Determine which tag to use in the rewritten Token.
// If the rewriter is the string "nil", the tag will be empty.
// If the rewriter is non-empty, non-"nil", the tag will be the rewriter.
// If the rewriter is empty, the tag will be the supplied Token's tag.
// /*If the rewriter is a variable, the tag will be the supplied Token's tag.*/
//
#define RW(A,B) (((A)->rewriter)!="")?(((A)->rewriter)!="nil"?(A)->rewriter:""):(B)
//#define RW2(A,B) (((A)->rewriter)!=""/*&&((A)->rewriter)!="X"&&((A)->rewriter)!="x"&&((A)->rewriter)!="Y"&&((A)->rewriter)!="y"*/)?(((A)->rewriter)!="nil"?(A)->rewriter:""):(B)
#define RW2(A,B) ((A)->rewriter)=="X" || ((A)->rewriter)=="Y"

Token* Symbol::rewrite(Token* t) const
{
  string fu  = RW(funcs,t->func);
  if(funcs->rewriter=="Y") fu = t->sfunc;
  string fo  = RW(forms,t->form);
  if(forms->rewriter=="y") fo = t->sform;
  string sfu = RW(sfuncs,t->sfunc);
  string sfo = RW(sforms,t->sform);
  return t->doRewrite(fu,fo,sfu,sfo,t->word,t->base);
}

bool Symbol::isPunctuation() const
{
  return (wordforms->first())[0] == '$';
}

bool Symbol::isEmpty() const
{
  return funcs->isEmpty()&&forms->isEmpty()&&sfuncs->isEmpty()&&sforms->isEmpty()&&wordforms->isEmpty()&&baseforms->isEmpty();
}
