#ifndef _Token_hh_
#define _Token_hh_

#include <string>
#include <vector>
#include <list>

class Symbol;

class Token
{
private:
  Token * prev;
  
protected:
  virtual void doSynonyms(bool);
  
public:
  static unsigned int cloneInvoked, appendMatchesInvoked, constructorInvoked;
  
  string func;
  string form;
  string sfunc;
  string sform;
  string word;
  string base;
  list<Token*>* syns;
  
  virtual unsigned int getSize() const { return 1; }
  virtual bool hasImpossibleDiscontinuities() const { return false; }
  virtual Token* clone() const;
  virtual string toVISL(int level=0, bool useClosedCorpusFormat = false) const;
  virtual string toString() const;
  virtual bool matches(Symbol *) const;
  virtual vector<Token*>* appendMatches(Symbol * sym, vector<Token*>* v=0) const;
  virtual Token* doRewrite(string fu, string fo, string sfu, string sfo, string w, string b) const;
  virtual bool isPunctuation() const;
  
  virtual void    setPrevious(Token *);
  virtual Token * getPrevious() const;
  
  virtual int getLeafCount() const { return 1; }
  
  Token(string fu="",  string fo="", 
	string sfu="", string sfo="", 
	string w="",   string b="",

	bool syn=true) :
    func(fu), form(fo),
    sfunc(sfu), sform(sfo),
    word(w), base(b)
    {
      prev = 0;
      ++constructorInvoked;
      doSynonyms(syn);
    }
  
  virtual ~Token()
    {}
};

class TokenGroup : public Token
{
 public:
  vector<Token*>* children;
  
  virtual Token* clone() const;
  virtual string toVISL(int level=0, bool useClosedCorpusFormat = false) const;
  virtual string toString() const;
  virtual Token* doRewrite(string fu, string fo, string sfu, string sfo, string w, string d) const;
  virtual unsigned int getSize() const { return children->size(); };
  virtual bool hasImpossibleDiscontinuities() const;
  
  virtual int getLeafCount() const;
  
  TokenGroup() : Token() { children = new vector<Token*>(); }
  TokenGroup(vector<Token*>* set,
	     string fu="",  string fo="", 
	     string sfu="", string sfo="", 
	     string w="",   string b="",
	     bool syn=true) :
    Token(fu,fo,sfu,sfo,w,b,syn),
    children(set)
    {}
  
  virtual ~TokenGroup()
    {}
};

class TokenSet : public TokenGroup
{
 public:
  virtual unsigned int getSize() const { return 1; }
  virtual vector<Token*>* appendMatches(Symbol * sym, vector<Token*>* v=0) const;
  virtual bool matches(Symbol * sym) const;
  virtual bool hasImpossibleDiscontinuities() const { return false; }
  
  virtual string toVISL(int level=0, bool useClosedCorpusFormat = false) const;
  virtual string toString() const;  
  virtual Token* doRewrite(string fu, string fo, string sfu, string sfo, string w, string d) const;
  
  virtual int getLeafCount() const { return 1; }
  
  TokenSet() : TokenGroup() {}
  TokenSet(vector<Token*>* set,
	   string fu="",  string fo="", 
	   string sfu="", string sfo="", 
	   string w="",   string b="",
	   bool syn=true) :
    TokenGroup(set,fu,fo,sfu,sfo,w,b,syn)
    {}
  
  virtual ~TokenSet()
    {}
};

#endif
